/////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
//
// Realtime Tropical Island Generator v1.0 by Vladeta Stojanovic (062920@live.abertay.ac.uk).
// Based on the OpenGL framework code provided by Matt Bett (m.bett@abertay.ac.uk).
// Additional code based on examples from the "OpenGL SuperBible, 5th Edition".
// Procedural generation code makes use of the LibNoise library (http://libnoise.sourceforge.net/).
//
// Version 1.0 (060511)
//
// Recorded bugs & fixes:
//				 - Camera z-axis slipping due to the use of Euler angles for performing local rotations.
//				   This has been fixed by performing the rotations along the X axis locally, while performing rotations around the Y axis using world
//				   matrix transformations. This way, there is no Gimbal lock, as the the X and Y rotations are never parallel within the same coordinates 
//				   set.
//
// Additional notes:
//               - This was my first attempt at real time procedural modeling. Thus, this code is not optimized for performance 
//				   and there are a lot of things that could have been/could still be done better and in a more efficient manner.
//				   The procedural generation makes use of LibNoise. LibNoise, in my opinion, is not ideally optimized for real
//				   time application, but is sufficient for demonstrative purpuses. Therefore, the procedural generation at run time
//				   of each island can take up to 45 seconds. This is one of the things to be adressed and potentially optimized 
//				   in the future releases of the application.
//
//				 - The rendering of mesh data is not very efficient, since the use of VBO's and geometry shaders was omited due to 
//				   time constraints. As a result, 90% of the rendering computations (with the exception of the water mesh animation which is done via
//				   the vertex shader on the GPU), are done on CPU rather then on the GPU. This includes the procedural generation of textures and mesh
//				   data. For the future releases of the application (or any application that I will be working on based on the key procedural modeling
//				   principals covered in example application), all of the procedural data generation should be moved to the GPU for more efficient computation.
//
//////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////



#include <cassert>
#include <iostream>
#include "NeHeWindow.h"
#include "directinput.h"
#include "igame.h"
#include "GL/glut.h"		//always include last.  can throw up compile errors otherwise
#include "tofromstring.h"
#include "IInput.h"
#include "Camera.h"

//Includes for the OpenGL toolkit.
#include <GLTools.h>	
#include <GLMatrixStack.h>
#include <GLFrame.h>
#include <GLFrustum.h>
#include <GLBatch.h>
#include <GLGeometryTransform.h>
#include <math3d.h>
#include "texloader.h"
#include "TestTerrain.h"
#include "TreeClass.h"
#include "CoherentNoise.h"
#include "Water.h"
#include "SkyBox.h"

using namespace std;

#define MAXTREES 1024 //over 32^2 and your CPU might die...

GLShaderManager		shaderManager;
GLGeometryTransform	transformPipeline;
GLGeometryTransform	treePipeline;
GLGeometryTransform waterPipeline;
GLGeometryTransform skyPipeline;

///////////////////////
// Shader variables
//////////////////////
GLuint	ADSTextureShader;	//Terrain shader
GLint   skyBoxShader;	//Skybox shader
GLint	waterShader;	//Water shader

//Terrain shader specific
GLint	locAmbient;			// The location of the ambient color
GLint   locDiffuse;			// The location of the diffuse color
GLint	locLight;			// The location of the Light in eye coordinates
GLint	locMVP;				// The location of the ModelViewProjection matrix uniform
GLint	locMV;				// The location of the ModelView matrix uniform
GLint	locNM;				// The location of the Normal matrix uniform
GLint   locTexture;

//Skybox specific
GLint	locMVPSkyBox;

//Water shader specific
GLint	locMVPWater;
GLint	locWaterAmbient;		
GLint   locWaterDiffuse;		
GLint	locWaterLight;			
GLint	locWaterMVP;			
GLint	locWaterMV;				
GLint	locWaterNM;				
GLint   locWaterTexture;
GLfloat	locTime;
GLfloat locWaveSpeed;
float waveTime;

//Added by vlad
GLMatrixStack			modelViewMatrix;	
GLMatrixStack			TreeMVMatrix;
GLMatrixStack			WaterMVMatrix;		
GLMatrixStack			SkyMVMatrix;	
GLMatrixStack			projectionMatrix;	
GLFrustum				viewFrustum;	
Camera					g_Camera;
TestTerrain				g_Terrain;
PalmTrees				gTrees;
Water					gWater;
SkyBox					gSkybox;
GLBatch					cubeBatch;

//Test textures
GLuint terrainTexture;
GLuint waterTexture;
GLuint cubeTexture;

//Skybox sides (note that only a single texture is generated and used for all 6 sides)
const char *szCubeFaces[6] = { "../data/skyTex.bmp", "../data/skyTex.bmp", "../data/skyTex.bmp", "../data/skyTex.bmp", "../data/skyTex.bmp", "../data/skyTex.bmp" };
GLenum  cube[6] = {  GL_TEXTURE_CUBE_MAP_POSITIVE_X,
                     GL_TEXTURE_CUBE_MAP_NEGATIVE_X,
                     GL_TEXTURE_CUBE_MAP_POSITIVE_Y,
                     GL_TEXTURE_CUBE_MAP_NEGATIVE_Y,
                     GL_TEXTURE_CUBE_MAP_POSITIVE_Z,
                     GL_TEXTURE_CUBE_MAP_NEGATIVE_Z };

//Material values
static GLfloat vWhite[] = { 1.0f, 1.0f, 1.0f, 1.0f };
static GLfloat vLeafColor[] = {0.0f, 1.0f, 0.0f, 1.0f};
static GLfloat vTrunkColor[] = {0.5f, 0.25f, 0.0f, 1.0f};

//Lighting data (for the terrain)
static GLfloat lightAmbient[]	= { 0.2f, 0.2f, 0.2f, 1.0f };
static GLfloat lightDiffuse[]	= { 1.0f, 1.0f, 1.0f, 1.0f };

//Light vector data
M3DVector4f vLightPos	= { 256.0f, 200.0f, 256.0f, 1.0f  };
M3DVector4f vLightEyePos;

static bool treeValid[1024];

// The 'Game'  Class
class ShaderApp : public IGame 
{
public:
	ShaderApp(std::tr1::shared_ptr<IInput> pinput);

	virtual	bool OnStartUp();

	virtual void OnShutDown();

	virtual void Update(const float dtSeconds);

	virtual void OnWindowResized( const int newWidth, const int newHeight );

	virtual void OnWindowMoved(const int newX, const int newY);

	std::tr1::shared_ptr<IInput> input_;

private:

	void CalcTreePlacement();

};


//Functions for the Object Factory

std::tr1::shared_ptr<IWindow> get_game_window( std::tr1::shared_ptr<IGame> game, 
                                               std::tr1::shared_ptr<IInput> input, 
                                               const void* pWindowHandle )
{
	return std::tr1::shared_ptr<IWindow>(new NeHeWindow( game, input, *reinterpret_cast<const HINSTANCE*>(pWindowHandle)));
}


std::tr1::shared_ptr<IGame> get_game(std::tr1::shared_ptr<IInput> input)
{
	return std::tr1::shared_ptr<IGame>(new ShaderApp(input));
}


std::tr1::shared_ptr<IInput> get_game_input()
{
	return std::tr1::shared_ptr<IInput>(new DirectInput());
}


///////////////////////////////////////////////////////////////////////////////////
ShaderApp::ShaderApp(std::tr1::shared_ptr<IInput> pinput):
IGame(pinput)
{
	input_ = pinput;
}

void ShaderApp::Update(const float dtSeconds)
{	
	////////////////////////////////////
	// Setup the scene and the camera
	////////////////////////////////////
	glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT | GL_STENCIL_BUFFER_BIT);

	float height = g_Terrain.GetYFieldVertValue(g_Camera.GetX(), g_Camera.GetZ());
	g_Camera.FPSWalk(height, 1.0f);
	g_Camera.UpdateCamera();
	
	//Wave timer for the water
	waveTime += 0.01f;

	//Setup the lighting here
	m3dTransformVector4(vLightEyePos, vLightPos, g_Camera.mCamera);

	///////////////////////
	//Render the skybox
	///////////////////////
	g_Camera.m_cameraFrame.GetCameraMatrix(g_Camera.mCamera, false);

	SkyMVMatrix.PushMatrix();
	SkyMVMatrix.LoadIdentity();
	SkyMVMatrix.PushMatrix(g_Camera.mCamera);
	SkyMVMatrix.Translate(g_Camera.GetX(), g_Camera.GetY(), g_Camera.GetZ());

	SkyMVMatrix.PushMatrix();

		glBindTexture(GL_TEXTURE_CUBE_MAP, cubeTexture);
		glUseProgram(skyBoxShader);
		glUniformMatrix4fv(locMVPSkyBox, 1, GL_FALSE, skyPipeline.GetModelViewProjectionMatrix());
		glDisable(GL_DEPTH_TEST);
			cubeBatch.Draw();      
		glEnable(GL_DEPTH_TEST);

	  SkyMVMatrix.PopMatrix();
    SkyMVMatrix.PopMatrix();
	SkyMVMatrix.PopMatrix();

	////////////////////////
	//Render the terrain
	////////////////////////
	modelViewMatrix.PushMatrix();	
	modelViewMatrix.LoadIdentity();
	modelViewMatrix.PushMatrix(g_Camera.mCamera);

	glBindTexture(GL_TEXTURE_2D, terrainTexture);
	glUseProgram(ADSTextureShader);
	glUniform4fv(locAmbient, 1, lightAmbient);
	glUniform4fv(locDiffuse, 1, lightDiffuse);
	glUniform3fv(locLight, 1, vLightEyePos);
	glUniformMatrix4fv(locMVP, 1, GL_FALSE, transformPipeline.GetModelViewProjectionMatrix());
	glUniformMatrix4fv(locMV, 1, GL_FALSE, transformPipeline.GetModelViewMatrix());
	glUniformMatrix3fv(locNM, 1, GL_FALSE, transformPipeline.GetNormalMatrix());
	glUniform1i(locTexture, 0);

		g_Terrain.Render();

	modelViewMatrix.PopMatrix();
	modelViewMatrix.PopMatrix();

	////////////////////
	//Draw the water
	////////////////////
	WaterMVMatrix.PushMatrix();	
	WaterMVMatrix.LoadIdentity();
	WaterMVMatrix.PushMatrix(g_Camera.mCamera);

	WaterMVMatrix.Translate(0.0f, 1.9f, 0.0f);

	//Setup a basic diffuse texture shader that will be used for drawing the terrain
	glBindTexture(GL_TEXTURE_2D,  waterTexture);
	glUseProgram(waterShader);
	glUniform4fv(locWaterAmbient, 1, lightAmbient);
	glUniform4fv(locWaterDiffuse, 1, lightDiffuse);
	glUniform3fv(locWaterLight, 1, vLightEyePos);
	glUniformMatrix4fv(locWaterMVP, 1, GL_FALSE, waterPipeline.GetModelViewProjectionMatrix());
	glUniformMatrix4fv(locWaterMV, 1, GL_FALSE, waterPipeline.GetModelViewMatrix());
	glUniformMatrix3fv(locWaterNM, 1, GL_FALSE,	waterPipeline.GetNormalMatrix());
	glUniform1f(locTime, waveTime);
	glUniform1f(locWaveSpeed, 15.0);
	glUniform1i(locWaterTexture, 0);

		gWater.RenderWater();

	WaterMVMatrix.PopMatrix();
	WaterMVMatrix.PopMatrix();

	///////////////////
	//Draw the trees 
	///////////////////

	TreeMVMatrix.PushMatrix();
	TreeMVMatrix.LoadIdentity();
	TreeMVMatrix.PushMatrix(g_Camera.mCamera);

	glDisable(GL_CULL_FACE);

	//First the palm leaves
	for(int i = 0; i < MAXTREES ; ++i)
	{
		TreeMVMatrix.PushMatrix();
		TreeMVMatrix.Translate(gTrees.mX[i], gTrees.mY[i], gTrees.mZ[i]);
		TreeMVMatrix.Scale(0.25f, 0.25f, 0.25f);

		shaderManager.UseStockShader(GLT_SHADER_POINT_LIGHT_DIFF, treePipeline.GetModelViewMatrix(), 
							treePipeline.GetProjectionMatrix(), vLightEyePos, vLeafColor);
		if(treeValid[i])
		{
			gTrees.DrawTreeLeaf();
		}

		TreeMVMatrix.PopMatrix();
	}

	//Then the trunk
	for(int j = 0; j < MAXTREES ; ++j)
	{
		TreeMVMatrix.PushMatrix();
		TreeMVMatrix.Translate(gTrees.mX[j], gTrees.mY[j], gTrees.mZ[j]);
		TreeMVMatrix.Scale(0.25f, 0.25f, 0.25f);

		shaderManager.UseStockShader(GLT_SHADER_POINT_LIGHT_DIFF, treePipeline.GetModelViewMatrix(), 
							treePipeline.GetProjectionMatrix(), vLightEyePos, vTrunkColor);
		if(treeValid[j])
		{
			gTrees.DrawTreeTrunk();
		}

		TreeMVMatrix.PopMatrix();
	}

	glEnable(GL_CULL_FACE);

	TreeMVMatrix.PopMatrix();
	TreeMVMatrix.PopMatrix();

	///////////////////////////////////////////////////////////////
	// If user presses SPACE, then regenerate new random island.
	///////////////////////////////////////////////////////////////
	if(GetKeyState(VK_SPACE) & 0x80)
	{
		gSkybox.GenerateSkybox();
		gSkybox.RenderCubeFaces();
		
		glBindTexture(GL_TEXTURE_CUBE_MAP, cubeTexture);
		for(int i = 0; i < 6; i++)
		{       
			LoadBMPCubeTexture(szCubeFaces[i], cube[i]); 
		}
    
		g_Terrain.ReGenerate();
		g_Terrain.Init();
		glBindTexture(GL_TEXTURE_2D, terrainTexture);
		LoadBMPTexture("../data/terrainTex.bmp", GL_LINEAR_MIPMAP_LINEAR, GL_LINEAR, GL_REPEAT); 

		gWater.GenerateWaterTex();
		glBindTexture(GL_TEXTURE_2D, waterTexture);
		LoadBMPTexture("../data/waterTex.bmp", GL_LINEAR_MIPMAP_LINEAR, GL_LINEAR, GL_REPEAT); 

		CalcTreePlacement();
	}

	if(GetKeyState('V') & 0x80)
	{
		//Render everything wireframe
		glPolygonMode(GL_FRONT_AND_BACK, GL_LINE);
	}
	else if(GetKeyState('C') & 0x80)
	{
		//Go back to normal
		glPolygonMode(GL_FRONT_AND_BACK, GL_FILL);
	}
}

void ShaderApp::CalcTreePlacement()
{
	gTrees.PlaceTreeBatch(0.01f);

	for(int i = 0; i < 1024; ++i)
	{
		if(g_Terrain.GetYFieldVertValue(gTrees.mX[i],gTrees.mZ[i])  > 14) //current elevation restriction is 12. Trees placed on any point in the map below the
		{																  //value of 12 won't get rendered.
			treeValid[i] = true;
		}
		else
		{
			treeValid[i] = false;
		}
	}
}

bool ShaderApp::OnStartUp()
{
	int argcp = 0;
    char* argv = "";
    glutInit( &argcp, &argv );  

	shaderManager.InitializeStockShaders();

	glEnable(GL_DEPTH_TEST);
	glEnable(GL_CULL_FACE);
	glFrontFace(GL_CW);
	glPolygonMode(GL_FRONT, GL_FILL);
	glCullFace(GL_BACK);

	//Setup cam
	g_Camera.InitCam();

	//Skybox shader
	skyBoxShader = gltLoadShaderPairWithAttributes("../data/SkyBox.vp", "../data/SkyBox.fp", 2, 
                                                GLT_ATTRIBUTE_VERTEX, "vVertex",
                                                GLT_ATTRIBUTE_NORMAL, "vNormal");

	locMVPSkyBox = glGetUniformLocation(skyBoxShader, "mvpMatrix");

	//Phong shader
	ADSTextureShader = gltLoadShaderPairWithAttributes(	"../data/ADSTexture.vp", 
														"../data/ADSTexture.fp", 
													3, 
													GLT_ATTRIBUTE_VERTEX, "vVertex",
													GLT_ATTRIBUTE_NORMAL, "vNormal", 
													GLT_ATTRIBUTE_TEXTURE0, "vTexture0");
	//setup all ourvariables to set uniform values.
	locAmbient = glGetUniformLocation(ADSTextureShader, "ambientColor");
	locDiffuse = glGetUniformLocation(ADSTextureShader, "diffuseColor");
	locLight = glGetUniformLocation(ADSTextureShader, "vLightPosition");
	locMVP = glGetUniformLocation(ADSTextureShader, "mvpMatrix");
	locMV  = glGetUniformLocation(ADSTextureShader, "mvMatrix");
	locNM  = glGetUniformLocation(ADSTextureShader, "normalMatrix");
	locTexture = glGetUniformLocation(ADSTextureShader, "colorMap");

	//Water shader
	waterShader =  gltLoadShaderPairWithAttributes("../data/Water.vp", "../data/Water.fp", 3, 
                                                GLT_ATTRIBUTE_VERTEX, "vVertex",
                                                GLT_ATTRIBUTE_NORMAL, "vNormal",
												GLT_ATTRIBUTE_TEXTURE0, "vTexture0");
	locWaterAmbient = glGetUniformLocation(waterShader, "ambientColor");
	locWaterDiffuse = glGetUniformLocation(waterShader, "diffuseColor");
	locWaterLight = glGetUniformLocation(waterShader, "vLightPosition");
	locWaterMVP = glGetUniformLocation(waterShader, "mvpMatrix");
	locWaterMV  = glGetUniformLocation(waterShader, "mvMatrix");
	locWaterNM  = glGetUniformLocation(waterShader, "normalMatrix");
	locWaterTexture = glGetUniformLocation(waterShader, "colorMap");
	locTime = glGetUniformLocation(waterShader, "fTime");
	locWaveSpeed = glGetUniformLocation(waterShader, "fWaveSpeed");

	//Setup the skybox
	gSkybox.GenerateSkybox();
	gSkybox.RenderCubeFaces();

	//Setup terrain
	g_Terrain.ReGenerate();
	g_Terrain.Init();

	//setup the water
	gWater.InitWater();
	gWater.GenerateWaterTex();

	//Setup the trees
	CalcTreePlacement();

	//Generate texture id's
	glGenTextures(1, &cubeTexture);
	glGenTextures(1, &terrainTexture);
	glGenTextures(1, &waterTexture);

	//Setup terrain texturing
	glBindTexture(GL_TEXTURE_2D, terrainTexture);
	LoadBMPTexture("../data/terrainTex.bmp", GL_LINEAR_MIPMAP_LINEAR, GL_LINEAR, GL_REPEAT); 

	//Setup water texturing
	glBindTexture(GL_TEXTURE_2D,  waterTexture);
	LoadBMPTexture("../data/waterTex.bmp", GL_LINEAR_MIPMAP_LINEAR, GL_LINEAR, GL_REPEAT); 

	// Load cube map images
	glBindTexture(GL_TEXTURE_CUBE_MAP, cubeTexture);

    for(int i = 0; i < 6; i++)
    {       
       LoadBMPCubeTexture(szCubeFaces[i], cube[i]); 
    }
    
    gltMakeCube(cubeBatch, 20.0f);

	return true;
}

void ShaderApp::OnShutDown()
{	
	//Add any cleanup code here
	glDeleteTextures(1, &terrainTexture);
	glDeleteTextures(1, &waterTexture);
	glDeleteTextures(1, &cubeTexture);

	//also remember to delete all the data files generated by the application
	remove("../data/terrainTex.bmp");
	remove("../data/waterTex.bmp");
	remove("../data/skyTex.bmp");
}

void ShaderApp::OnWindowResized( const int newWidth, const int newHeight )
{
	// Create the projection matrix, and load it on the projection matrix stack
	viewFrustum.SetPerspective(35.0f, (float)newWidth/(float)newHeight, 1.0f, 1000.0f);
	projectionMatrix.LoadMatrix(viewFrustum.GetProjectionMatrix());
    
    // Set the transformation pipeline to use the desired matrix stacks 
	transformPipeline.SetMatrixStacks(modelViewMatrix, projectionMatrix);
	waterPipeline.SetMatrixStacks(WaterMVMatrix, projectionMatrix);
	treePipeline.SetMatrixStacks(TreeMVMatrix,  projectionMatrix);
	skyPipeline.SetMatrixStacks(SkyMVMatrix, projectionMatrix);
}

void ShaderApp::OnWindowMoved(const int newX, const int newY)
{

}
