#include <cassert>
#include <exception>
#include <windows.h>
#include <mmsystem.h>
#include "iwindow.h"
#include "igame.h"
#include "iinput.h"


// Constructs the window and ensures the user has given us a valid input and game
// to work with.
IWindow::IWindow( std::tr1::shared_ptr<IGame> game, 
                  std::tr1::shared_ptr<IInput> input ) 
    :   game_(game), 
        input_(input), 
        update_previously_called_(false)
{
    if( ! game_ )
    {
        throw new std::exception( "IGame pointer cannot be null" );
    }

    if( ! input_ )
    {
        throw new std::exception( "IInput pointer cannot be null" );
    }
}


// Informs the game it is shutting down, so long as the game was
// started to begin with.
IWindow::~IWindow()
{
    if( update_previously_called_ )
    {
        game_->OnShutDown();
    }
}


// This should be called every frame by the implementation the IWindow. This
// call is in turn responsible for updating the input and the game objects.
// It also keeps track of the time delta. 
void IWindow::UpdateFrame()
{
    input_->Update();

    if( update_previously_called_ )
    {
	    assert( game_ );
        unsigned long dtMs = timer_.deltaMs(); 

        if( dtMs > 0 )
        {
            // We only call the game window when there
            // has been at least a millisecond difference
            // since the last call. This guarantees to the 
            // game that it is never called with a timestep
            // of zero. 
            const float dtSeconds = dtMs * 0.0001f;
	        game_->Update( dtSeconds );
        }
    }
    else 
    {
        // To ensure we track time accurately, and also to 
        // allow the game to perform any special initialisation
        // we call startup first. This happens only once.
        timer_.start();
        update_previously_called_ = true;
        game_->OnStartUp();
    }
}


void IWindow::WindowResized( const int newWidth, const int newHeight )
{
	game_->OnWindowResized( newWidth, newHeight );
}

void IWindow::WindowMoved(const int newX, const int newY)
{
	game_->OnWindowMoved(newX, newY);
}