#define NOMINMAX
#include <exception>
#include <algorithm>
#include <Windows.h>
#include <mmsystem.h>
#include "timer.h"


// See http://msdn.microsoft.com/en-us/library/dd757624%28VS.85%29.aspx 
// for details of the usage of the timeBeginPeriod calls. They must always
// be paired with a timeEndPeriod. We guarantee this here.


timer::timer() : timestamp_(0), fidelityMs_(0)
{
    startup();
}

timer::~timer()
{
    shutdown();
}

timer::timer(const timer& t) : timestamp_(t.timestamp_)
{
    startup();
}

// We pair every call of startup with a corresponding call to shutdown
// by tying the calls to the lifetime of the objects.
void timer::startup()
{
    const UINT resolution = 1;
    TIMECAPS tc;

    if ( ::timeGetDevCaps(&tc, sizeof(TIMECAPS)) != TIMERR_NOERROR) 
    {
        throw std::exception( "Unable to startup timer." );
    }

    const UINT wTimerRes = std::min( std::max(tc.wPeriodMin, resolution), tc.wPeriodMax);
    ::timeBeginPeriod(wTimerRes); 

    fidelityMs_ = wTimerRes;
}

void timer::shutdown()
{
    timeEndPeriod(static_cast<UINT>(fidelityMs_));
}

void timer::start()
{
    timestamp_ = ::timeGetTime();
    timestamp_ = 0 == timestamp_ ? 1 : timestamp_; // Don't allow zero timestamp.
}

void timer::stop()
{
    timestamp_ = 0;
}

unsigned long timer::deltaMs()
{
    unsigned long dtMs = 0;

    // We are using a zero timestamp to also see if the timer is on/off. This
    // results in a potential loss of 1ms accuracy every ~50 days or so.
    // See http://msdn.microsoft.com/en-us/library/dd757629%28VS.85%29.aspx

    if( timestamp_ > 0 )
    {
        unsigned long timenow = ::timeGetTime();
        dtMs = timenow - timestamp_;
        if( dtMs > 0 )
        {
            timestamp_ = (timenow > 0) ? timenow : 1;
        }
    }

    return dtMs;
    
}